'use strict';

var sander = require('sander');
var path = require('path');
var MagicString = require('magic-string');
MagicString = 'default' in MagicString ? MagicString['default'] : MagicString;
var acorn = require('acorn');

function unixizePath(path$$) {
    if (path.sep === "/") return path$$;
    return path$$.split(path.sep).join("/");
}

function getIndentString(magicString, options) {
	if (!('indent' in options) || options.indent === true) {
		return magicString.getIndentString();
	}

	return options.indent || '';
}

function badExports(option, keys) {
	throw new Error('\'' + option + '\' was specified for options.exports, but entry module has following exports: ' + keys.join(', '));
}

var keys = Object.keys;

function getExportMode(bundle, exportMode) {
	var exportKeys = keys(bundle.entryModule.exports);

	if (exportMode === 'default') {
		if (exportKeys.length !== 1 || exportKeys[0] !== 'default') {
			badExports('default', exportKeys);
		}
	} else if (exportMode === 'none' && exportKeys.length) {
		badExports('none', exportKeys);
	}

	if (!exportMode || exportMode === 'auto') {
		if (exportKeys.length === 0) {
			exportMode = 'none';
		} else if (exportKeys.length === 1 && exportKeys[0] === 'default') {
			exportMode = 'default';
		} else {
			exportMode = 'named';
		}
	}

	if (!/(?:default|named|none)/.test(exportMode)) {
		throw new Error('options.exports must be \'default\', \'named\', \'none\', \'auto\', or left unspecified (defaults to \'auto\')');
	}

	return exportMode;
}

function getName(x) {
	return x.name;
}

function req(x) {
	return "require('" + x.id + "')";
}

function quoteId(x) {
	return "'" + x.id + "'";
}

function blank() {
	return Object.create(null);
}

function umd(bundle, magicString, _ref, options) {
	var exportMode = _ref.exportMode;
	var indentString = _ref.indentString;

	if (exportMode !== 'none' && !options.moduleName) {
		throw new Error('You must supply options.moduleName for UMD bundles');
	}

	var globalNames = options.globals || blank();

	var amdDeps = bundle.externalModules.map(quoteId);
	var cjsDeps = bundle.externalModules.map(req);
	var globalDeps = bundle.externalModules.map(function (module) {
		return globalNames[module.id] || module.name;
	});

	var args = bundle.externalModules.map(getName);

	if (exportMode === 'named') {
		amdDeps.unshift('\'exports\'');
		cjsDeps.unshift('exports');
		globalDeps.unshift('(global.' + options.moduleName + ' = {})');

		args.unshift('exports');
	}

	var amdParams = (options.moduleId ? '[\'' + options.moduleId + '\'], ' : '') + (amdDeps.length ? '[' + amdDeps.join(', ') + '], ' : '');

	var cjsExport = exportMode === 'default' ? 'module.exports = ' : '';
	var defaultExport = exportMode === 'default' ? 'global.' + options.moduleName + ' = ' : '';

	var intro = ('(function (global, factory) {\n\t\t\ttypeof exports === \'object\' && typeof module !== \'undefined\' ? ' + cjsExport + 'factory(' + cjsDeps.join(', ') + ') :\n\t\t\ttypeof define === \'function\' && define.amd ? define(' + amdParams + 'factory) :\n\t\t\t' + defaultExport + 'factory(' + globalDeps + ');\n\t\t}(this, function (' + args + ') { \'use strict\';\n\n\t\t').replace(/^\t\t/gm, '').replace(/^\t/gm, magicString.getIndentString());

	var exports = bundle.entryModule.exports;

	var exportBlock = undefined;

	if (exportMode === 'default') {
		var canonicalName = bundle.entryModule.getCanonicalName('default');
		exportBlock = 'return ' + canonicalName + ';';
	} else {
		exportBlock = bundle.toExport.map(function (name) {
			var canonicalName = bundle.entryModule.getCanonicalName(exports[name].localName);
			return 'exports.' + name + ' = ' + canonicalName + ';';
		}).join('\n');
	}

	if (exportBlock) {
		magicString.append('\n\n' + exportBlock);
	}

	return magicString.trim().indent(indentString).append('\n\n}));').prepend(intro);
}

function iife(bundle, magicString, _ref, options) {
	var exportMode = _ref.exportMode;
	var indentString = _ref.indentString;

	var globalNames = options.globals || blank();

	var dependencies = bundle.externalModules.map(function (module) {
		return globalNames[module.id] || module.name;
	});

	var args = bundle.externalModules.map(getName);

	if (exportMode !== 'none' && !options.moduleName) {
		throw new Error('You must supply options.moduleName for IIFE bundles');
	}

	if (exportMode === 'named') {
		dependencies.unshift('(this.' + options.moduleName + ' = {})');
		args.unshift('exports');
	}

	var intro = '(function (' + args + ') { \'use strict\';\n\n';
	var outro = '\n\n})(' + dependencies + ');';

	if (exportMode === 'default') {
		intro = 'var ' + options.moduleName + ' = ' + intro;
		magicString.append('\n\nreturn ' + bundle.entryModule.getCanonicalName('default') + ';');
	}

	// TODO named exports

	return magicString.indent(indentString).prepend(intro).append(outro);
}

function es6(bundle, magicString, _ref, options) {
	var exportMode = _ref.exportMode;

	var introBlock = ''; // TODO...

	var exports = bundle.entryModule.exports;
	var exportBlock = keys(exports).map(function (exportedName) {
		var specifier = exports[exportedName];

		var canonicalName = bundle.entryModule.getCanonicalName(specifier.localName);

		if (exportedName === 'default') {
			return 'export default ' + canonicalName + ';';
		}

		return exportedName === canonicalName ? 'export { ' + exportedName + ' };' : 'export { ' + canonicalName + ' as ' + exportedName + ' };';
	}).join('\n');

	if (exportBlock) {
		magicString.append('\n\n' + exportBlock);
	}

	return magicString.trim();
}

function cjs(bundle, magicString, _ref) {
	var exportMode = _ref.exportMode;

	var intro = '\'use strict\';\n\n';

	// TODO handle empty imports, once they're supported
	var importBlock = bundle.externalModules.map(function (module) {
		var requireStatement = 'var ' + module.name + ' = require(\'' + module.id + '\');';

		if (module.needsDefault) {
			requireStatement += '\n' + (module.needsNamed ? 'var ' + module.name + '__default = ' : '' + module.name + ' = ') + ('\'default\' in ' + module.name + ' ? ' + module.name + '[\'default\'] : ' + module.name + ';');
		}

		return requireStatement;
	}).join('\n');

	if (importBlock) {
		intro += importBlock + '\n\n';
	}

	magicString.prepend(intro);

	var exportBlock = undefined;
	if (exportMode === 'default' && bundle.entryModule.exports.default) {
		exportBlock = 'module.exports = ' + bundle.entryModule.getCanonicalName('default') + ';';
	} else if (exportMode === 'named') {
		exportBlock = bundle.toExport.map(function (key) {
			var specifier = bundle.entryModule.exports[key];
			var name = bundle.entryModule.getCanonicalName(specifier.localName);

			return 'exports.' + key + ' = ' + name + ';';
		}).join('\n');
	}

	if (exportBlock) {
		magicString.append('\n\n' + exportBlock);
	}

	return magicString;
}

function amd(bundle, magicString, _ref, options) {
	var exportMode = _ref.exportMode;
	var indentString = _ref.indentString;

	var deps = bundle.externalModules.map(quoteId);
	var args = bundle.externalModules.map(getName);

	if (exportMode === 'named') {
		args.unshift('exports');
		deps.unshift('\'exports\'');
	}

	var params = (options.moduleId ? '[\'' + options.moduleId + '\'], ' : '') + (deps.length ? '[' + deps.join(', ') + '], ' : '');

	var intro = 'define(' + params + 'function (' + args.join(', ') + ') { \'use strict\';\n\n';

	var exports = bundle.entryModule.exports;

	var exportBlock = undefined;

	if (exportMode === 'default') {
		exportBlock = 'return ' + bundle.entryModule.getCanonicalName('default') + ';';
	} else {
		exportBlock = bundle.toExport.map(function (name) {
			return 'exports.' + name + ' = ' + exports[name].localName + ';';
		}).join('\n');
	}

	if (exportBlock) {
		magicString.append('\n\n' + exportBlock);
	}

	return magicString.indent(indentString).append('\n\n});').prepend(intro);
}

var finalisers = { amd: amd, cjs: cjs, es6: es6, iife: iife, umd: umd };

var blacklisted = blank();

var builtins = 'Infinity NaN undefined null true false eval uneval isFinite isNaN parseFloat parseInt decodeURI decodeURIComponent encodeURI encodeURIComponent escape unescape Object Function Boolean Symbol Error EvalError InternalError RangeError ReferenceError SyntaxError TypeError URIError Number Math Date String RegExp Array Int8Array Uint8Array Uint8ClampedArray Int16Array Uint16Array Int32Array Uint32Array Float32Array Float64Array Map Set WeakMap WeakSet SIMD ArrayBuffer DataView JSON Promise Generator GeneratorFunction Reflect Proxy Intl'.split(' ');

var reservedWords = 'break case class catch const continue debugger default delete do else export extends finally for function if import in instanceof let new return super switch this throw try typeof var void while with yield enum await implements package protected static interface private public'.split(' ');

reservedWords.concat(builtins).forEach(function (word) {
	return blacklisted[word] = true;
});
function makeLegalIdentifier(str) {
	str = str.replace(/[^$_a-zA-Z0-9]/g, '_');
	if (/\d/.test(str[0]) || blacklisted[str]) str = '_' + str;

	return str;
}

function deconflict(name, names) {
	while (name in names) {
		name = '_' + name;
	}

	return name;
}

function sequence(arr, callback) {
	var len = arr.length;
	var results = new Array(len);

	var promise = sander.Promise.resolve();

	function next(i) {
		return promise.then(function () {
			return callback(arr[i], i);
		}).then(function (result) {
			return results[i] = result;
		});
	}

	var i = undefined;

	for (i = 0; i < len; i += 1) {
		promise = next(i);
	}

	return promise.then(function () {
		return results;
	});
}

var emptyArrayPromise = sander.Promise.resolve([]);

function analyse(magicString, module) {
	// first we need to generate comprehensive scope info
	var previousStatement = null;
	var commentIndex = 0;

	module.statements.forEach(function (statement) {
		var node = statement.node;

		var trailing = !!previousStatement;
		var previousComment = undefined;

		// TODO surely this can be neater
		// attach leading comment
		do {
			var comment = module.comments[commentIndex];

			// prevent comments inside the previous statement being
			// appended to it
			if (previousStatement) {
				while (comment && comment.start < previousStatement.node.end) {
					commentIndex += 1;
					comment = module.comments[commentIndex];
				}
			}

			if (!comment || comment.end > node.start) break;

			// attach any trailing comment to the previous statement
			if (trailing && !/\n/.test(module.source.slice(previousStatement.node.end, comment.start))) {
				previousStatement.trailingComment = comment;
			}

			// then attach leading comments to this statement
			else {
				statement.leadingComments.push({
					separator: previousComment ? magicString.slice(previousComment.end, comment.start) : '\n',
					comment: comment
				});

				previousComment = comment;
			}

			commentIndex += 1;
			trailing = false;
		} while (module.comments[commentIndex]);

		// determine margin
		var previousEnd = previousComment ? previousComment.end : previousStatement ? (previousStatement.trailingComment || previousStatement.node).end : 0;

		//const start = ( statement.leadingComments[0] || node ).start;

		var gap = magicString.original.slice(previousEnd, node.start);
		var margin = gap.split('\n').length;

		if (previousStatement) previousStatement.margin[1] = margin;
		statement.margin[0] = margin;

		statement.analyse();

		previousStatement = statement;
	});
}

var _analyse = analyse;

function getLocation(source, charIndex) {
	var lines = source.split('\n');
	var len = lines.length;

	var lineStart = 0;
	var i = undefined;

	for (i = 0; i < len; i += 1) {
		var line = lines[i];
		var lineEnd = lineStart + line.length + 1; // +1 for newline

		if (lineEnd > charIndex) {
			return { line: i + 1, column: charIndex - lineStart };
		}

		lineStart = lineEnd;
	}

	throw new Error('Could not determine location of character');
}

function isExportDeclaration(statement) {
	return statement.isExportDeclaration;
}

function isImportDeclaration(statement) {
	return statement.isImportDeclaration;
}

var shouldAbort = undefined;

var shouldSkip = undefined;

var context = {
	skip: function () {
		return shouldSkip = true;
	},
	abort: function () {
		return shouldAbort = true;
	}
};

var toString = Object.prototype.toString;

function isArray(thing) {
	return toString.call(thing) === '[object Array]';
}

var childKeys = blank();

function visit(node, parent, enter, leave) {
	if (!node || shouldAbort) return;

	if (enter) {
		shouldSkip = false;
		enter.call(context, node, parent);
		if (shouldSkip || shouldAbort) return;
	}

	var keys = childKeys[node.type] || (childKeys[node.type] = Object.keys(node).filter(function (key) {
		return typeof node[key] === 'object';
	}));

	var key = undefined,
	    value = undefined,
	    i = undefined,
	    j = undefined;

	i = keys.length;
	while (i--) {
		key = keys[i];
		value = node[key];

		if (isArray(value)) {
			j = value.length;
			while (j--) {
				visit(value[j], node, enter, leave);
			}
		} else if (value && value.type) {
			visit(value, node, enter, leave);
		}
	}

	if (leave && !shouldAbort) {
		leave(node, parent);
	}
}

function walk(ast, _ref) {
	var enter = _ref.enter;
	var leave = _ref.leave;

	shouldAbort = false;
	visit(ast, null, enter, leave);
}

var blockDeclarations = {
	'const': true,
	'let': true
};

function __classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var Scope = (function () {
	function Scope(options) {
		var _this = this;

		__classCallCheck(this, Scope);

		options = options || {};

		this.parent = options.parent;
		this.depth = this.parent ? this.parent.depth + 1 : 0;
		this.declarations = blank();
		this.isBlockScope = !!options.block;

		if (options.params) {
			options.params.forEach(function (param) {
				_this.declarations[param.name] = param;
			});
		}
	}

	// add ( name, isBlockDeclaration ) {
	// 	if ( !isBlockDeclaration && this.isBlockScope ) {
	// 		// it's a `var` or function declaration, and this
	// 		// is a block scope, so we need to go up
	// 		this.parent.add( name, isBlockDeclaration );
	// 	} else {
	// 		this.names.push( name );
	// 	}
	// }

	Scope.prototype.addDeclaration = function addDeclaration(name, declaration) {
		var isBlockDeclaration = declaration.type === 'VariableDeclaration' && blockDeclarations[declaration.kind];

		if (!isBlockDeclaration && this.isBlockScope) {
			// it's a `var` or function declaration, and this
			// is a block scope, so we need to go up
			this.parent.addDeclaration(name, declaration);
		} else {
			this.declarations[name] = declaration;
		}
	};

	Scope.prototype.getDeclaration = function getDeclaration(name) {
		return this.declarations[name] || this.parent && this.parent.getDeclaration(name);
	};

	Scope.prototype.contains = function contains(name) {
		return !!this.getDeclaration(name);
	};

	Scope.prototype.findDefiningScope = function findDefiningScope(name) {
		if (!!this.declarations[name]) {
			return this;
		}

		if (this.parent) {
			return this.parent.findDefiningScope(name);
		}

		return null;
	};

	return Scope;
})();

function ___classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var Statement = (function () {
	function Statement(node, magicString, module, index) {
		___classCallCheck(this, Statement);

		this.node = node;
		this.module = module;
		this.magicString = magicString;
		this.index = index;
		this.id = module.id + '#' + index;

		this.scope = new Scope();
		this.defines = blank();
		this.modifies = blank();
		this.dependsOn = blank();
		this.stronglyDependsOn = blank();

		this.isIncluded = false;

		this.leadingComments = [];
		this.trailingComment = null;
		this.margin = [0, 0];

		// some facts about this statement...
		this.isImportDeclaration = node.type === 'ImportDeclaration';
		this.isExportDeclaration = /^Export/.test(node.type);
	}

	Statement.prototype.analyse = function analyse() {
		var _this = this;

		if (this.isImportDeclaration) return; // nothing to analyse

		var statement = this; // TODO use arrow functions instead
		var magicString = this.magicString;

		var scope = this.scope;

		walk(this.node, {
			enter: function (node) {
				var newScope = undefined;

				magicString.addSourcemapLocation(node.start);

				switch (node.type) {
					case 'FunctionExpression':
					case 'FunctionDeclaration':
					case 'ArrowFunctionExpression':
						if (node.type === 'FunctionDeclaration') {
							scope.addDeclaration(node.id.name, node);
						}

						newScope = new Scope({
							parent: scope,
							params: node.params, // TODO rest params?
							block: false
						});

						// named function expressions - the name is considered
						// part of the function's scope
						if (node.type === 'FunctionExpression' && node.id) {
							newScope.addDeclaration(node.id.name, node);
						}

						break;

					case 'BlockStatement':
						newScope = new Scope({
							parent: scope,
							block: true
						});

						break;

					case 'CatchClause':
						newScope = new Scope({
							parent: scope,
							params: [node.param],
							block: true
						});

						break;

					case 'VariableDeclaration':
						node.declarations.forEach(function (declarator) {
							scope.addDeclaration(declarator.id.name, node);
						});
						break;

					case 'ClassDeclaration':
						scope.addDeclaration(node.id.name, node);
						break;
				}

				if (newScope) {
					Object.defineProperty(node, '_scope', { value: newScope });
					scope = newScope;
				}
			},
			leave: function (node) {
				if (node._scope) {
					scope = scope.parent;
				}
			}
		});

		if (!this.isImportDeclaration) {
			walk(this.node, {
				enter: function (node, parent) {
					if (node._scope) scope = node._scope;

					_this.checkForReads(scope, node, parent);
					_this.checkForWrites(scope, node);
				},
				leave: function (node) {
					if (node._scope) scope = scope.parent;
				}
			});
		}

		keys(scope.declarations).forEach(function (name) {
			statement.defines[name] = true;
		});
	};

	Statement.prototype.checkForReads = function checkForReads(scope, node, parent) {
		if (node.type === 'Identifier') {
			// disregard the `bar` in `foo.bar` - these appear as Identifier nodes
			if (parent.type === 'MemberExpression' && node !== parent.object) {
				return;
			}

			// disregard the `bar` in { bar: foo }
			if (parent.type === 'Property' && node !== parent.value) {
				return;
			}

			// disregard the `bar` in `class Foo { bar () {...} }`
			if (parent.type === 'MethodDefinition') return;

			var definingScope = scope.findDefiningScope(node.name);

			if ((!definingScope || definingScope.depth === 0) && !this.defines[node.name]) {
				this.dependsOn[node.name] = true;

				if (!scope.parent) this.stronglyDependsOn[node.name] = true;
			}
		}
	};

	Statement.prototype.checkForWrites = function checkForWrites(scope, node) {
		var _this2 = this;

		var addNode = function (node, isAssignment) {
			var depth = 0; // determine whether we're illegally modifying a binding or namespace

			while (node.type === 'MemberExpression') {
				node = node.object;
				depth += 1;
			}

			// disallow assignments/updates to imported bindings and namespaces
			if (isAssignment) {
				var importSpecifier = _this2.module.imports[node.name];

				if (importSpecifier && !scope.contains(node.name)) {
					var minDepth = importSpecifier.name === '*' ? 2 : // cannot do e.g. `namespace.foo = bar`
					1; // cannot do e.g. `foo = bar`, but `foo.bar = bar` is fine

					if (depth < minDepth) {
						var err = new Error('Illegal reassignment to import \'' + node.name + '\'');
						err.file = _this2.module.id;
						err.loc = getLocation(_this2.module.magicString.toString(), node.start);
						throw err;
					}
				}

				// special case = `export default foo; foo += 1;` - we'll
				// need to assign a new variable so that the exported
				// value is not updated by the second statement
				if (_this2.module.exports.default && depth === 0 && _this2.module.exports.default.identifier === node.name) {
					// but only if this is a) inside a function body or
					// b) after the export declaration
					if (!!scope.parent || node.start > _this2.module.exports.default.statement.node.start) {
						_this2.module.exports.default.isModified = true;
					}
				}
			}

			if (node.type === 'Identifier') {
				_this2.modifies[node.name] = true;
			}
		};

		if (node.type === 'AssignmentExpression') {
			addNode(node.left, true);
		} else if (node.type === 'UpdateExpression') {
			addNode(node.argument, true);
		} else if (node.type === 'CallExpression') {
			node.arguments.forEach(function (arg) {
				return addNode(arg, false);
			});

			// `foo.bar()` is assumed to mutate foo
			if (node.callee.type === 'MemberExpression') {
				addNode(node.callee);
			}
		}
	};

	Statement.prototype.expand = function expand() {
		var _this3 = this;

		this.isIncluded = true; // prevent statement being included twice

		var result = [];

		// We have a statement, and it hasn't been included yet. First, include
		// the statements it depends on
		var dependencies = Object.keys(this.dependsOn);

		return sequence(dependencies, function (name) {
			if (_this3.defines[name]) return; // TODO maybe exclude from `this.dependsOn` in the first place?

			return _this3.module.define(name).then(function (definition) {
				result.push.apply(result, definition);
			});
		})

		// then include the statement itself
		.then(function () {
			result.push(_this3);
		})

		// then include any statements that could modify the
		// thing(s) this statement defines
		.then(function () {
			return sequence(keys(_this3.defines), function (name) {
				var modifications = _this3.module.modifications[name];

				if (modifications) {
					return sequence(modifications, function (statement) {
						if (!statement.isIncluded) {
							return statement.expand().then(function (statements) {
								result.push.apply(result, statements);
							});
						}
					});
				}
			});
		})

		// the `result` is an array of all statements that need
		// to be included if this one is
		.then(function () {
			return result;
		});
	};

	Statement.prototype.replaceIdentifiers = function replaceIdentifiers(names, bundleExports) {
		var magicString = this.magicString.clone();
		var replacementStack = [names];
		var nameList = keys(names);

		var deshadowList = [];
		nameList.forEach(function (name) {
			var replacement = names[name];
			deshadowList.push(replacement.split('.')[0]);
		});

		var topLevel = true;
		var depth = 0;

		walk(this.node, {
			enter: function (node, parent) {
				var _this4 = this;

				if (node._skip) return this.skip();

				if (/^Function/.test(node.type)) depth += 1;

				// `this` is undefined at the top level of ES6 modules
				if (node.type === 'ThisExpression' && depth === 0) {
					magicString.overwrite(node.start, node.end, 'undefined');
				}

				// special case - variable declarations that need to be rewritten
				// as bundle exports
				if (topLevel) {
					if (node.type === 'VariableDeclaration') {
						// if this contains a single declarator, and it's one that
						// needs to be rewritten, we replace the whole lot
						var _name = node.declarations[0].id.name;
						if (node.declarations.length === 1 && bundleExports[_name]) {
							magicString.overwrite(node.start, node.declarations[0].id.end, bundleExports[_name]);
							node.declarations[0].id._skip = true;
						}

						// otherwise, we insert the `exports.foo = foo` after the declaration
						else {
							var exportInitialisers = node.declarations.map(function (declarator) {
								return declarator.id.name;
							}).filter(function (name) {
								return !!bundleExports[name];
							}).map(function (name) {
								return '\n' + bundleExports[name] + ' = ' + name + ';';
							}).join('');

							// TODO clean this up
							try {
								magicString.insert(node.end, exportInitialisers);
							} catch (err) {
								magicString.append(exportInitialisers);
							}
						}
					}
				}

				var scope = node._scope;

				if (scope) {
					var _ret = (function () {
						topLevel = false;

						var newNames = blank();
						var hasReplacements = undefined;

						// special case = function foo ( foo ) {...}
						if (node.id && names[node.id.name] && scope.declarations[node.id.name]) {
							magicString.overwrite(node.id.start, node.id.end, names[node.id.name]);
						}

						keys(names).forEach(function (name) {
							if (!scope.declarations[name]) {
								newNames[name] = names[name];
								hasReplacements = true;
							}
						});

						deshadowList.forEach(function (name) {
							if (~scope.declarations[name]) {
								// TODO is this right? no indexOf?
								newNames[name] = name + '$$'; // TODO better mechanism
								hasReplacements = true;
							}
						});

						if (!hasReplacements && depth > 0) {
							return {
								v: _this4.skip()
							};
						}

						names = newNames;
						replacementStack.push(newNames);
					})();

					if (typeof _ret === 'object') return _ret.v;
				}

				// We want to rewrite identifiers (that aren't property names etc)
				if (node.type !== 'Identifier') return;
				if (parent.type === 'MemberExpression' && !parent.computed && node !== parent.object) return;
				if (parent.type === 'Property' && node !== parent.value) return;
				// TODO others...?

				var name = names[node.name];

				if (name && name !== node.name) {
					magicString.overwrite(node.start, node.end, name);
				}
			},

			leave: function (node) {
				if (/^Function/.test(node.type)) depth -= 1;

				if (node._scope) {
					replacementStack.pop();
					names = replacementStack[replacementStack.length - 1];
				}
			}
		});

		return magicString;
	};

	return Statement;
})();

function ____classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var Module = (function () {
	function Module(_ref) {
		var _this = this;

		var id = _ref.id;
		var source = _ref.source;
		var bundle = _ref.bundle;

		____classCallCheck(this, Module);

		this.source = source;

		this.bundle = bundle;
		this.id = id;

		// By default, `id` is the filename. Custom resolvers and loaders
		// can change that, but it makes sense to use it for the source filename
		this.magicString = new MagicString(source, {
			filename: id
		});

		this.suggestedNames = blank();
		this.comments = [];

		// Try to extract a list of top-level statements/declarations. If
		// the parse fails, attach file info and abort
		var ast = undefined;

		try {
			ast = acorn.parse(source, {
				ecmaVersion: 6,
				sourceType: 'module',
				onComment: function (block, text, start, end) {
					return _this.comments.push({ block: block, text: text, start: start, end: end });
				}
			});
		} catch (err) {
			err.code = 'PARSE_ERROR';
			err.file = id; // see above - not necessarily true, but true enough
			throw err;
		}

		walk(ast, {
			enter: function (node) {
				_this.magicString.addSourcemapLocation(node.start);
				_this.magicString.addSourcemapLocation(node.end);
			}
		});

		this.statements = [];

		ast.body.map(function (node) {
			// special case - top-level var declarations with multiple declarators
			// should be split up. Otherwise, we may end up including code we
			// don't need, just because an unwanted declarator is included
			if (node.type === 'VariableDeclaration' && node.declarations.length > 1) {
				node.declarations.forEach(function (declarator) {
					var magicString = _this.magicString.snip(declarator.start, declarator.end).trim();
					magicString.prepend('' + node.kind + ' ').append(';');

					var syntheticNode = {
						type: 'VariableDeclaration',
						kind: node.kind,
						start: node.start,
						end: node.end,
						declarations: [declarator]
					};

					var statement = new Statement(syntheticNode, magicString, _this, _this.statements.length);
					_this.statements.push(statement);
				});
			} else {
				var magicString = _this.magicString.snip(node.start, node.end).trim();
				var statement = new Statement(node, magicString, _this, _this.statements.length);

				_this.statements.push(statement);
			}
		});

		this.importDeclarations = this.statements.filter(isImportDeclaration);
		this.exportDeclarations = this.statements.filter(isExportDeclaration);

		this.analyse();
	}

	Module.prototype.analyse = function analyse() {
		var _this2 = this;

		// imports and exports, indexed by ID
		this.imports = blank();
		this.exports = blank();

		this.importDeclarations.forEach(function (statement) {
			var node = statement.node;
			var source = node.source.value;

			node.specifiers.forEach(function (specifier) {
				var isDefault = specifier.type === 'ImportDefaultSpecifier';
				var isNamespace = specifier.type === 'ImportNamespaceSpecifier';

				var localName = specifier.local.name;
				var name = isDefault ? 'default' : isNamespace ? '*' : specifier.imported.name;

				if (_this2.imports[localName]) {
					var err = new Error('Duplicated import \'' + localName + '\'');
					err.file = _this2.id;
					err.loc = getLocation(_this2.source, specifier.start);
					throw err;
				}

				_this2.imports[localName] = {
					source: source,
					name: name,
					localName: localName
				};
			});
		});

		this.exportDeclarations.forEach(function (statement) {
			var node = statement.node;
			var source = node.source && node.source.value;

			// export default function foo () {}
			// export default foo;
			// export default 42;
			if (node.type === 'ExportDefaultDeclaration') {
				var isDeclaration = /Declaration$/.test(node.declaration.type);
				var isAnonymous = /(?:Class|Function)Expression$/.test(node.declaration.type);

				var declaredName = isDeclaration && node.declaration.id.name;
				var identifier = node.declaration.type === 'Identifier' && node.declaration.name;

				_this2.exports.default = {
					statement: statement,
					name: 'default',
					localName: declaredName || 'default',
					declaredName: declaredName,
					identifier: identifier,
					isDeclaration: isDeclaration,
					isAnonymous: isAnonymous,
					isModified: false // in case of `export default foo; foo = somethingElse`
				};
			}

			// export { foo, bar, baz }
			// export var foo = 42;
			// export function foo () {}
			else if (node.type === 'ExportNamedDeclaration') {
				if (node.specifiers.length) {
					// export { foo, bar, baz }
					node.specifiers.forEach(function (specifier) {
						var localName = specifier.local.name;
						var exportedName = specifier.exported.name;

						_this2.exports[exportedName] = {
							localName: localName,
							exportedName: exportedName
						};

						// export { foo } from './foo';
						if (source) {
							_this2.imports[localName] = {
								source: source,
								localName: localName,
								name: localName
							};
						}
					});
				} else {
					var declaration = node.declaration;

					var _name = undefined;

					if (declaration.type === 'VariableDeclaration') {
						// export var foo = 42
						_name = declaration.declarations[0].id.name;
					} else {
						// export function foo () {}
						_name = declaration.id.name;
					}

					_this2.exports[_name] = {
						statement: statement,
						localName: _name,
						expression: declaration
					};
				}
			}
		});

		_analyse(this.magicString, this);

		this.canonicalNames = blank();

		this.definitions = blank();
		this.definitionPromises = blank();
		this.modifications = blank();

		this.statements.forEach(function (statement) {
			keys(statement.defines).forEach(function (name) {
				_this2.definitions[name] = statement;
			});

			keys(statement.modifies).forEach(function (name) {
				if (!_this2.modifications[name]) {
					_this2.modifications[name] = [];
				}

				_this2.modifications[name].push(statement);
			});
		});

		this.statements.forEach(function (statement) {
			keys(statement.dependsOn).forEach(function (name) {
				if (!_this2.definitions[name] && !_this2.imports[name]) {
					_this2.bundle.assumedGlobals[name] = true;
				}
			});
		});
	};

	Module.prototype.findDeclaration = function findDeclaration(localName) {
		var importDeclaration = this.imports[localName];

		// name was defined by another module
		if (importDeclaration) {
			var _module = importDeclaration.module;

			if (_module.isExternal) return null;

			var exportDeclaration = _module.exports[importDeclaration.name];
			return _module.findDeclaration(exportDeclaration.localName);
		}

		// name was defined by this module, if any
		var i = this.statements.length;
		while (i--) {
			var declaration = this.statements[i].scope.declarations[localName];
			if (declaration) {
				return declaration;
			}
		}

		return null;
	};

	Module.prototype.getCanonicalName = function getCanonicalName(localName) {
		// Special case
		if (localName === 'default' && (this.exports.default.isModified || !this.suggestedNames.default)) {
			var canonicalName = makeLegalIdentifier(this.id.replace(path.dirname(this.bundle.entryModule.id) + '/', '').replace(/\.js$/, ''));
			return deconflict(canonicalName, this.definitions);
		}

		if (this.suggestedNames[localName]) {
			localName = this.suggestedNames[localName];
		}

		if (!this.canonicalNames[localName]) {
			var canonicalName = undefined;

			if (this.imports[localName]) {
				var importDeclaration = this.imports[localName];
				var _module2 = importDeclaration.module;

				if (importDeclaration.name === '*') {
					canonicalName = _module2.suggestedNames['*'];
				} else {
					var exporterLocalName = undefined;

					if (_module2.isExternal) {
						exporterLocalName = importDeclaration.name;
					} else {
						var exportDeclaration = _module2.exports[importDeclaration.name];
						exporterLocalName = exportDeclaration.localName;
					}

					canonicalName = _module2.getCanonicalName(exporterLocalName);
				}
			} else {
				canonicalName = localName;
			}

			this.canonicalNames[localName] = canonicalName;
		}

		return this.canonicalNames[localName];
	};

	Module.prototype.define = function define(name) {
		var _this3 = this;

		// shortcut cycles. TODO this won't work everywhere...
		if (this.definitionPromises[name]) {
			return emptyArrayPromise;
		}

		var promise = undefined;

		// The definition for this name is in a different module
		if (this.imports[name]) {
			(function () {
				var importDeclaration = _this3.imports[name];

				promise = _this3.bundle.fetchModule(importDeclaration.source, _this3.id).then(function (module) {
					importDeclaration.module = module;

					// suggest names. TODO should this apply to non default/* imports?
					if (importDeclaration.name === 'default') {
						// TODO this seems ropey
						var localName = importDeclaration.localName;
						var suggestion = _this3.suggestedNames[localName] || localName;

						// special case - the module has its own import by this name
						while (!module.isExternal && module.imports[suggestion]) {
							suggestion = '_' + suggestion;
						}

						module.suggestName('default', suggestion);
					} else if (importDeclaration.name === '*') {
						var localName = importDeclaration.localName;
						var suggestion = _this3.suggestedNames[localName] || localName;
						module.suggestName('*', suggestion);
						module.suggestName('default', '' + suggestion + '__default');
					}

					if (module.isExternal) {
						if (importDeclaration.name === 'default') {
							module.needsDefault = true;
						} else {
							module.needsNamed = true;
						}

						module.importedByBundle.push(importDeclaration);
						return emptyArrayPromise;
					}

					if (importDeclaration.name === '*') {
						// we need to create an internal namespace
						if (! ~_this3.bundle.internalNamespaceModules.indexOf(module)) {
							_this3.bundle.internalNamespaceModules.push(module);
						}

						return module.expandAllStatements();
					}

					var exportDeclaration = module.exports[importDeclaration.name];

					if (!exportDeclaration) {
						throw new Error('Module ' + module.id + ' does not export ' + importDeclaration.name + ' (imported by ' + _this3.id + ')');
					}

					return module.define(exportDeclaration.localName);
				});
			})();
		}

		// The definition is in this module
		else if (name === 'default' && this.exports.default.isDeclaration) {
			// We have something like `export default foo` - so we just start again,
			// searching for `foo` instead of default
			promise = this.define(this.exports.default.name);
		} else {
			(function () {
				var statement = undefined;

				statement = name === 'default' ? _this3.exports.default.statement : _this3.definitions[name];
				promise = statement && !statement.isIncluded ? statement.expand() : emptyArrayPromise;

				// Special case - `export default foo; foo += 1` - need to be
				// vigilant about maintaining the correct order of the export
				// declaration. Otherwise, the export declaration will always
				// go at the end of the expansion, because the expansion of
				// `foo` will include statements *after* the declaration
				if (name === 'default' && _this3.exports.default.identifier && _this3.exports.default.isModified) {
					(function () {
						var defaultExportStatement = _this3.exports.default.statement;
						promise = promise.then(function (statements) {
							// remove the default export statement...
							// TODO could this be statements.pop()?
							statements.splice(statements.indexOf(defaultExportStatement), 1);

							var i = statements.length;
							var inserted = false;

							while (i--) {
								if (statements[i].module === _this3 && statements[i].index < defaultExportStatement.index) {
									statements.splice(i + 1, 0, defaultExportStatement);
									inserted = true;
									break;
								}
							}

							if (!inserted) statements.push(statement);
							return statements;
						});
					})();
				}
			})();
		}

		this.definitionPromises[name] = promise || emptyArrayPromise;
		return this.definitionPromises[name];
	};

	Module.prototype.expandAllStatements = function expandAllStatements(isEntryModule) {
		var _this4 = this;

		var allStatements = [];

		return sequence(this.statements, function (statement) {
			// A statement may have already been included, in which case we need to
			// curb rollup's enthusiasm and move it down here. It remains to be seen
			// if this approach is bulletproof
			if (statement.isIncluded) {
				var index = allStatements.indexOf(statement);
				if (~index) {
					allStatements.splice(index, 1);
					allStatements.push(statement);
				}

				return;
			}

			// skip import declarations...
			if (statement.isImportDeclaration) {
				// ...unless they're empty, in which case assume we're importing them for the side-effects
				// THIS IS NOT FOOLPROOF. Probably need /*rollup: include */ or similar
				if (!statement.node.specifiers.length) {
					return _this4.bundle.fetchModule(statement.node.source.value, _this4.id).then(function (module) {
						statement.module = module;
						return module.expandAllStatements();
					}).then(function (statements) {
						allStatements.push.apply(allStatements, statements);
					});
				}

				return;
			}

			// skip `export { foo, bar, baz }`...
			if (statement.node.type === 'ExportNamedDeclaration' && statement.node.specifiers.length) {
				// ...but ensure they are defined, if this is the entry module
				if (isEntryModule) {
					return statement.expand().then(function (statements) {
						allStatements.push.apply(allStatements, statements);
					});
				}

				return;
			}

			// include everything else
			return statement.expand().then(function (statements) {
				allStatements.push.apply(allStatements, statements);
			});
		}).then(function () {
			return allStatements;
		});
	};

	Module.prototype.rename = function rename(name, replacement) {
		this.canonicalNames[name] = replacement;
	};

	Module.prototype.suggestName = function suggestName(defaultOrBatch, suggestion) {
		// deconflict anonymous default exports with this module's definitions
		var shouldDeconflict = this.exports.default && this.exports.default.isAnonymous;

		if (shouldDeconflict) suggestion = deconflict(suggestion, this.definitions);

		if (!this.suggestedNames[defaultOrBatch]) {
			this.suggestedNames[defaultOrBatch] = makeLegalIdentifier(suggestion);
		}
	};

	return Module;
})();

function _____classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var ExternalModule = (function () {
	function ExternalModule(id) {
		_____classCallCheck(this, ExternalModule);

		this.id = id;
		this.name = null;

		this.isExternal = true;
		this.importedByBundle = [];

		this.canonicalNames = blank();
		this.suggestedNames = blank();

		this.needsDefault = false;
		this.needsNamed = false;
	}

	ExternalModule.prototype.getCanonicalName = function getCanonicalName(name) {
		if (name === 'default') {
			return this.needsNamed ? '' + this.name + '__default' : this.name;
		}

		if (name === '*') {
			return this.name;
		}

		// TODO this depends on the output format... works for CJS etc but not ES6
		return '' + this.name + '.' + name;
	};

	ExternalModule.prototype.rename = function rename(name, replacement) {
		this.canonicalNames[name] = replacement;
	};

	ExternalModule.prototype.suggestName = function suggestName(exportName, suggestion) {
		if (!this.suggestedNames[exportName]) {
			this.suggestedNames[exportName] = suggestion;
		}
	};

	ExternalModule.prototype.findDefiningStatement = function findDefiningStatement() {
		return null;
	};

	return ExternalModule;
})();

function ensureArray(thing) {
	if (Array.isArray(thing)) return thing;
	if (thing == undefined) return [];
	return [thing];
}

var absolutePath = /^(?:\/|(?:[A-Za-z]:)?\\)/;
function defaultResolver(importee, importer, options) {
	// absolute paths are left untouched
	if (absolutePath.test(importee)) return importee;

	// if this is the entry point, resolve against cwd
	if (importer === undefined) return path.resolve(importee);

	// we try to resolve external modules
	if (importee[0] !== '.') {
		// unless we want to keep it external, that is
		if (~options.external.indexOf(importee)) return null;

		return options.resolveExternal(importee, importer, options);
	}

	return path.resolve(path.dirname(importer), importee).replace(/\.js$/, '') + '.js';
}

function defaultExternalResolver(id, importer) {
	// for now, only node_modules is supported, and only jsnext:main
	var root = absolutePath.exec(importer)[0];
	var dir = path.dirname(importer);

	while (dir !== root) {
		var pkgPath = path.resolve(dir, 'node_modules', id, 'package.json');
		var pkgJson = undefined;

		try {
			pkgJson = sander.readFileSync(pkgPath).toString();
		} catch (err) {}

		if (pkgJson) {
			var pkg = undefined;

			try {
				pkg = JSON.parse(pkgJson);
			} catch (err) {
				throw new Error('Malformed JSON: ' + pkgPath);
			}

			var main = pkg['jsnext:main'];

			if (!main) {
				throw new Error('Package ' + id + ' does not have a jsnext:main field, and so cannot be included in your rollup. Try adding it as an external module instead (e.g. options.external = [\'' + id + '\']). See https://github.com/rollup/rollup/wiki/jsnext:main for more info');
			}

			return path.resolve(path.dirname(pkgPath), main).replace(/\.js$/, '') + '.js';
		}

		dir = path.dirname(dir);
	}

	throw new Error('Could not find package ' + id + ' (required by ' + importer + ')');
}

function defaultLoader(id, options) {
	// TODO support plugins e.g. !css and !json?
	var source = sander.readFileSync(id, { encoding: 'utf-8' });

	return options.transform.reduce(function (source, transformer) {
		return transformer(source, id);
	}, source);
}

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var Bundle = (function () {
	function Bundle(options) {
		_classCallCheck(this, Bundle);

		this.entry = options.entry;
		this.entryModule = null;

		this.resolveId = options.resolveId || defaultResolver;
		this.load = options.load || defaultLoader;

		this.resolveOptions = {
			external: ensureArray(options.external),
			resolveExternal: options.resolveExternal || defaultExternalResolver
		};

		this.loadOptions = {
			transform: ensureArray(options.transform)
		};

		this.varExports = blank();
		this.toExport = null;

		this.modulePromises = blank();
		this.statements = [];
		this.externalModules = [];
		this.internalNamespaceModules = [];
		this.assumedGlobals = blank();
	}

	Bundle.prototype.fetchModule = function fetchModule(importee, importer) {
		var _this = this;

		return sander.Promise.resolve(this.resolveId(importee, importer, this.resolveOptions)).then(function (id) {
			if (!id) {
				// external module
				if (!_this.modulePromises[importee]) {
					var _module = new ExternalModule(importee);
					_this.externalModules.push(_module);
					_this.modulePromises[importee] = sander.Promise.resolve(_module);
				}

				return _this.modulePromises[importee];
			}

			if (!_this.modulePromises[id]) {
				_this.modulePromises[id] = sander.Promise.resolve(_this.load(id, _this.loadOptions)).then(function (source) {
					var module = new Module({
						id: id,
						source: source,
						bundle: _this
					});

					return module;
				});
			}

			return _this.modulePromises[id];
		});
	};

	Bundle.prototype.build = function build() {
		var _this2 = this;

		// bring in top-level AST nodes from the entry module
		return this.fetchModule(this.entry, undefined).then(function (entryModule) {
			var defaultExport = entryModule.exports.default;

			_this2.entryModule = entryModule;

			if (defaultExport) {
				// `export default function foo () {...}` -
				// use the declared name for the export
				if (defaultExport.declaredName) {
					entryModule.suggestName('default', defaultExport.declaredName);
				}

				// `export default a + b` - generate an export name
				// based on the id of the entry module
				else {
					(function () {
						var defaultExportName = makeLegalIdentifier(path.basename(_this2.entryModule.id).slice(0, -path.extname(_this2.entryModule.id).length));

						// deconflict
						var topLevelNames = [];
						entryModule.statements.forEach(function (statement) {
							keys(statement.defines).forEach(function (name) {
								return topLevelNames.push(name);
							});
						});

						while (~topLevelNames.indexOf(defaultExportName)) {
							defaultExportName = '_' + defaultExportName;
						}

						entryModule.suggestName('default', defaultExportName);
					})();
				}
			}

			return entryModule.expandAllStatements(true);
		}).then(function (statements) {
			_this2.statements = statements;
			_this2.deconflict();
			_this2.sort();
		});
	};

	Bundle.prototype.deconflict = function deconflict() {
		var _this3 = this;

		var definers = blank();
		var conflicts = blank();

		// Discover conflicts (i.e. two statements in separate modules both define `foo`)
		this.statements.forEach(function (statement) {
			var module = statement.module;
			var names = keys(statement.defines);

			// with default exports that are expressions (`export default 42`),
			// we need to ensure that the name chosen for the expression does
			// not conflict
			if (statement.node.type === 'ExportDefaultDeclaration') {
				var _name = module.getCanonicalName('default');

				var isProxy = statement.node.declaration && statement.node.declaration.type === 'Identifier';
				var shouldDeconflict = !isProxy || module.getCanonicalName(statement.node.declaration.name) !== _name;

				if (shouldDeconflict && ! ~names.indexOf(_name)) {
					names.push(_name);
				}
			}

			names.forEach(function (name) {
				if (definers[name]) {
					conflicts[name] = true;
				} else {
					definers[name] = [];
				}

				// TODO in good js, there shouldn't be duplicate definitions
				// per module... but some people write bad js
				definers[name].push(module);
			});
		});

		// Assign names to external modules
		this.externalModules.forEach(function (module) {
			// TODO is this right?
			var name = makeLegalIdentifier(module.suggestedNames['*'] || module.suggestedNames.default || module.id);

			if (definers[name]) {
				conflicts[name] = true;
			} else {
				definers[name] = [];
			}

			definers[name].push(module);
			module.name = name;
		});

		// Ensure we don't conflict with globals
		keys(this.assumedGlobals).forEach(function (name) {
			if (definers[name]) {
				conflicts[name] = true;
			}
		});

		// Rename conflicting identifiers so they can live in the same scope
		keys(conflicts).forEach(function (name) {
			var modules = definers[name];

			if (!_this3.assumedGlobals[name]) {
				// the module closest to the entryModule gets away with
				// keeping things as they are, unless we have a conflict
				// with a global name
				modules.pop();
			}

			modules.forEach(function (module) {
				var replacement = getSafeName(name);
				module.rename(name, replacement);
			});
		});

		function getSafeName(name) {
			while (conflicts[name]) {
				name = '_' + name;
			}

			conflicts[name] = true;
			return name;
		}
	};

	Bundle.prototype.sort = function sort() {
		// TODO avoid this work whenever possible...

		var definitions = blank();

		// gather definitions
		this.statements.forEach(function (statement) {
			keys(statement.defines).forEach(function (name) {
				var canonicalName = statement.module.getCanonicalName(name);
				definitions[canonicalName] = statement;
			});
		});

		var strongDeps = blank();
		var stronglyDependsOn = blank();

		this.statements.forEach(function (statement) {
			var id = statement.id;
			strongDeps[id] = [];
			stronglyDependsOn[id] = {};

			keys(statement.stronglyDependsOn).forEach(function (name) {
				if (statement.defines[name]) return; // TODO seriously... need to fix this
				var canonicalName = statement.module.getCanonicalName(name);
				var definition = definitions[canonicalName];

				if (definition) strongDeps[statement.id].push(definition);
			});
		});

		// add second (and third...) order strong dependencies
		this.statements.forEach(function (statement) {
			var id = statement.id;

			// add second (and third...) order dependencies
			function addStrongDependencies(dependency) {
				if (stronglyDependsOn[id][dependency.id]) return;

				stronglyDependsOn[id][dependency.id] = true;
				strongDeps[dependency.id].forEach(addStrongDependencies);
			}

			strongDeps[id].forEach(addStrongDependencies);
		});

		// reinsert each statement, ensuring its strong dependencies appear first
		var sorted = [];
		var included = blank();

		function include(statement) {
			if (included[statement.id]) return;
			included[statement.id] = true;

			var len = sorted.length;
			var i = 0;
			var alreadyIncluded = false;

			for (i = 0; i < len; i += 1) {
				// ensure that this statement appears above later statements
				// from the same module - in rare situations (#34) they can
				// become jumbled
				var existing = sorted[i];
				if (existing.module === statement.module && existing.index > statement.index) {
					sorted.splice(i, 0, statement);
					alreadyIncluded = true;
				}
			}

			if (!alreadyIncluded) sorted.push(statement);
		}

		this.statements.forEach(function (statement) {
			strongDeps[statement.id].forEach(includeStrongDependency);

			function includeStrongDependency(dependency) {
				if (!stronglyDependsOn[dependency.id][statement.id] && !included[dependency.id]) {
					strongDeps[dependency.id].forEach(includeStrongDependency);
					include(dependency);
				}
			}

			include(statement);
		});

		this.statements = sorted;
	};

	Bundle.prototype.generate = function generate() {
		var _this4 = this;

		var options = arguments[0] === undefined ? {} : arguments[0];

		var magicString = new MagicString.Bundle({ separator: '' });

		var format = options.format || 'es6';

		// If we have named exports from the bundle, and those exports
		// are assigned to *within* the bundle, we may need to rewrite e.g.
		//
		//   export let count = 0;
		//   export function incr () { count++ }
		//
		// might become...
		//
		//   exports.count = 0;
		//   function incr () {
		//     exports.count += 1;
		//   }
		//   exports.incr = incr;
		//
		// This doesn't apply if the bundle is exported as ES6!
		var allBundleExports = blank();

		if (format !== 'es6') {
			keys(this.entryModule.exports).forEach(function (key) {
				var exportDeclaration = _this4.entryModule.exports[key];

				var originalDeclaration = _this4.entryModule.findDeclaration(exportDeclaration.localName);

				if (originalDeclaration && originalDeclaration.type === 'VariableDeclaration') {
					var canonicalName = _this4.entryModule.getCanonicalName(exportDeclaration.localName);

					allBundleExports[canonicalName] = 'exports.' + key;
					_this4.varExports[key] = true;
				}
			});
		}

		// since we're rewriting variable exports, we want to
		// ensure we don't try and export them again at the bottom
		this.toExport = keys(this.entryModule.exports).filter(function (key) {
			return !_this4.varExports[key];
		});

		// Apply new names and add to the output bundle
		var previousModule = null;
		var previousIndex = -1;
		var previousMargin = 0;

		this.statements.forEach(function (statement) {
			// skip `export { foo, bar, baz }`
			if (statement.node.type === 'ExportNamedDeclaration' && statement.node.specifiers.length) {
				return;
			}

			var replacements = blank();
			var bundleExports = blank();

			keys(statement.dependsOn).concat(keys(statement.defines)).forEach(function (name) {
				var canonicalName = statement.module.getCanonicalName(name);

				if (allBundleExports[canonicalName]) {
					bundleExports[name] = replacements[name] = allBundleExports[canonicalName];
				} else if (name !== canonicalName) {
					replacements[name] = canonicalName;
				}
			});

			var source = statement.replaceIdentifiers(replacements, bundleExports);

			// modify exports as necessary
			if (statement.isExportDeclaration) {
				// remove `export` from `export var foo = 42`
				if (statement.node.type === 'ExportNamedDeclaration' && statement.node.declaration.type === 'VariableDeclaration') {
					source.remove(statement.node.start, statement.node.declaration.start);
				}

				// remove `export` from `export class Foo {...}` or `export default Foo`
				// TODO default exports need different treatment
				else if (statement.node.declaration.id) {
					source.remove(statement.node.start, statement.node.declaration.start);
				} else if (statement.node.type === 'ExportDefaultDeclaration') {
					var _module2 = statement.module;
					var canonicalName = _module2.getCanonicalName('default');

					if (statement.node.declaration.type === 'Identifier' && canonicalName === _module2.getCanonicalName(statement.node.declaration.name)) {
						return;
					}

					// anonymous functions should be converted into declarations
					if (statement.node.declaration.type === 'FunctionExpression') {
						source.overwrite(statement.node.start, statement.node.declaration.start + 8, 'function ' + canonicalName);
					} else {
						source.overwrite(statement.node.start, statement.node.declaration.start, 'var ' + canonicalName + ' = ');
					}
				} else {
					throw new Error('Unhandled export');
				}
			}

			// ensure there is always a newline between statements, and add
			// additional newlines as necessary to reflect original source
			var minSeparation = previousModule !== statement.module || statement.index !== previousIndex + 1 ? 3 : 2;
			var margin = Math.max(minSeparation, statement.margin[0], previousMargin);
			var newLines = new Array(margin).join('\n');

			// add leading comments
			if (statement.leadingComments.length) {
				var commentBlock = newLines + statement.leadingComments.map(function (_ref) {
					var separator = _ref.separator;
					var comment = _ref.comment;

					return separator + (comment.block ? '/*' + comment.text + '*/' : '//' + comment.text);
				}).join('');

				magicString.addSource(new MagicString(commentBlock));
				newLines = new Array(statement.margin[0]).join('\n'); // TODO handle gaps between comment block and statement
			}

			// add the statement itself
			magicString.addSource({
				content: source,
				separator: newLines
			});

			// add trailing comments
			var comment = statement.trailingComment;
			if (comment) {
				var commentBlock = comment.block ? ' /*' + comment.text + '*/' : ' //' + comment.text;

				magicString.append(commentBlock);
			}

			previousMargin = statement.margin[1];
			previousModule = statement.module;
			previousIndex = statement.index;
		});

		// prepend bundle with internal namespaces
		var indentString = magicString.getIndentString();
		var namespaceBlock = this.internalNamespaceModules.map(function (module) {
			var exportKeys = keys(module.exports);

			return 'var ' + module.getCanonicalName('*') + ' = {\n' + exportKeys.map(function (key) {
				return '' + indentString + 'get ' + key + ' () { return ' + module.getCanonicalName(key) + '; }';
			}).join(',\n') + '\n};\n\n';
		}).join('');

		magicString.prepend(namespaceBlock);

		var finalise = finalisers[format];

		if (!finalise) {
			throw new Error('You must specify an output type - valid options are ' + keys(finalisers).join(', '));
		}

		magicString = finalise(this, magicString.trim(), {
			// Determine export mode - 'default', 'named', 'none'
			exportMode: getExportMode(this, options.exports),

			// Determine indentation
			indentString: getIndentString(magicString, options)
		}, options);

		var code = magicString.toString();
		var map = null;

		if (options.sourceMap) {
			(function () {
				map = magicString.generateMap({
					includeContent: true,
					file: options.sourceMapFile || options.dest
					// TODO
				});

				// make sources relative. TODO fix this upstream?
				var dir = path.dirname(map.file);
				map.sources = map.sources.map(function (source) {
					return source ? unixizePath(path.relative(dir, source)) : null;
				});
			})();
		}

		return { code: code, map: map };
	};

	return Bundle;
})();

var SOURCEMAPPING_URL = 'sourceMa';
SOURCEMAPPING_URL += 'ppingURL';
function rollup(options) {
	if (!options || !options.entry) {
		throw new Error('You must supply options.entry to rollup');
	}

	var bundle = new Bundle(options);

	return bundle.build().then(function () {
		return {
			generate: function (options) {
				return bundle.generate(options);
			},
			write: function (options) {
				if (!options || !options.dest) {
					throw new Error('You must supply options.dest to bundle.write');
				}

				var dest = options.dest;

				var _bundle$generate = bundle.generate(options);

				var code = _bundle$generate.code;
				var map = _bundle$generate.map;

				var promises = [];

				if (options.sourceMap) {
					var url = undefined;

					if (options.sourceMap === 'inline') {
						url = map.toUrl();
					} else {
						url = '' + path.basename(dest) + '.map';
						promises.push(sander.writeFile(dest + '.map', map.toString()));
					}

					code += '\n//# ' + SOURCEMAPPING_URL + '=' + url;
				}

				promises.push(sander.writeFile(dest, code));
				return Promise.all(promises);
			}
		};
	});
}

exports.rollup = rollup;